

#include "generalization.h"
#include "mech.h"
#include "mechDistortion.h"
#include "calCred.h"
#include "calViolation.h"
#include "ass.h"
#include "anatomy.h"

extern void transformToBinaryData(configure *config, equivalenceClass *eqClass, data *aData);
extern int *tclose_getDistFromEqTuple(equivalenceTuple *eqTuple, data *aData, int noOfSenValue);
extern double tclose_getSimilarValue(generalizedAllTable *genAllTable, int noOfDim, int noOfTuple, int noOfSenValue, int *dist1, int *dist2, int noOfRealElt1, int noOfRealElt2);

// 1: true
// 0: false
int isAlphaKAnonymity(data *dataSet, summary *synopsis, configure *config)
{
	int i;

	int noOfSen;

	int totalNo;

	//noOfSen = 0;

	// sorting according to the list of attributes in config->senAttrNo
	//data_sort(dataSet, config->noOfSenAttr, config->senAttrNo);
	//summary_sort(synopsis, config->noOfSenAttr, config->senAttrNo);

	// Assume only one sensitve attribute
	//int summary_getNoOfSen(data *dataSet, summary *synopsis, int noOfDim, int *dimNo)
	noOfSen = summary_getNoOfSen(dataSet, synopsis, config->noOfSenAttr, config->senAttrNo);

	// can be replaced with totalNo = dataSet->noOfTuple
	totalNo = 0;
	for (i = 0; i < synopsis->noOfGroup; i++)
	{
		totalNo += synopsis->grouping[i].noOfMember;
	}

	if (noOfSen/(double)totalNo <= config->alpha)
	{
		return 1;
	}
	else
	{
		return 0;
	}
	

}


void generalizeDimension(dimension *dim, generalizedLevelTable *genLevelTable)
{
	int i;

	int noOfPoss;


	noOfPoss = genLevelTable->noOfPoss;

	for (i = 0; i < noOfPoss; i++)
	{
		if (genLevelTable->gen[i].bottomValue == dim->value)
		{
			// generalize
			dim->value = genLevelTable->gen[i].topValue;
			dim->level = genLevelTable->gen[i].topLevel;

			break;
		}
	}
}

void generalizeEquivalenceTuple(equivalenceTuple *eqTuple, int attrNo, generalizedLevelTable *genLevelTable)
{
	dimension *dim;
	
	dim = &(eqTuple->content[attrNo]);
	generalizeDimension(dim, genLevelTable);
}

void generalizeEqClass(equivalenceClass *eqClass, int attrNo, configure *conf, generalizedAllTable *genAllTable)
{
	int i;

	int noOfDim;
	int genBottomLevel;

	equivalenceTuple **eqTuple;
	equivalenceTuple *eqOne;


	generalizedDimTable *genDimTable;
	generalizedLevelTable *genLevelTable;

	noOfDim = conf->noOfDim;

	genDimTable = &(genAllTable->genDimTable[attrNo]);

	eqTuple = (equivalenceTuple **) eqClass->eqClass; 

	if (eqClass->noOfEqClass > 0)
	{
		genBottomLevel = eqTuple[0]->content[attrNo].level;
		genLevelTable = &(genDimTable->gen[genBottomLevel]);

		for (i = 0; i < eqClass->noOfEqClass; i++)
		{
			//eqOne = &(eqTuple[i]);
			eqOne = eqTuple[i];
			generalizeEquivalenceTuple(eqOne, attrNo, genLevelTable);
		}
	}
	else
	{
		// do nothing
	}
}

void calculateSensitiveCount(equivalenceClass *eqClass, configure *conf, int *senCount, int *totalCount)
{
	int i;

	int inSenCount;
	int inTotalCount;

	equivalenceTuple **eqTuple;
	equivalenceTuple *eqOne;

	eqTuple = eqClass->eqClass;

	inSenCount = 0;
	inTotalCount = 0;
	for (i = 0; i < eqClass->noOfEqClass; i++)
	{
		//eqOne = &(eqTuple[i]);
		eqOne = eqTuple[i];

		inTotalCount += eqOne->noOfMember;
		inSenCount += eqOne->noOfSen;
	}

	(*senCount) = inSenCount;
	(*totalCount) = inTotalCount;
}

int calculateGreatestPossRemovalCount(equivalenceClass *eqClass, configure *conf)
{
	int returnValue;

	int senCount;
	int totalCount;

	calculateSensitiveCount(eqClass, conf, &senCount, &totalCount);

	// flooring
	returnValue = (int) (totalCount - senCount/conf->alpha);

	if (returnValue < 0)
	{
		returnValue = 0;
	}

	return returnValue;
}


void fullSpaceGeneralize(equivalenceClass *projEqClass, data *dataSet, configure *conf, generalizedAllTable *genAllTable, AttributeList *attrList, GeneralizationDomain *genDomain)
{
	int i, j;
	int noOfDim;

	int *list;
	int noOfAttr;

	int *genStrength;
	int *genLevel;

	int dimNo;

	noOfDim = conf->noOfDim-1;

	genLevel = genDomain->genLevel;

	// preprocess
	noOfAttr = attrList->noOfAttr;
	list = attrList->attr;

	genStrength = (int *) malloc(sizeof(int)*noOfDim);
	memset(genStrength, 0, sizeof(int)*noOfDim);
	for (i = 0; i < noOfAttr; i++)
	{
		dimNo = list[i];

		genStrength[dimNo] = genLevel[i];
	}
	

	// generalize
	for (i = 0; i < noOfDim; i++)
	{
		dimNo = i;
		for (j = 0; j < genStrength[dimNo]; j++)
		{
			generalizeEqClass(projEqClass, dimNo, conf, genAllTable);
		}
	}

	// merge
	mergeEqClass(projEqClass, dataSet, conf);

	free(genStrength);
}


int isSatisfyKAnonymity(equivalenceClass *eqClass, configure *conf)
{
	int noOfEqTuple;
	equivalenceTuple **eqTuple;

	int returnValue;

	equivalenceTuple *oneEqTuple;

	int noOfMember, noOfSen;

	double alpha;

	int K;

	int i;

	K = conf->K;

	noOfEqTuple = eqClass->noOfEqClass;
	eqTuple = eqClass->eqClass;

	alpha = conf->alpha;
	
	returnValue = 1;
	for (i = 0; i < noOfEqTuple; i++)
	{
		oneEqTuple = eqTuple[i];

		noOfMember = oneEqTuple->noOfMember;
		noOfSen = oneEqTuple->noOfSen;

		// *** IMPORTATN change for m-deassociation
		//if ((noOfMember < K) || (noOfSen > ceiling(alpha*noOfMember)))
		if ((noOfMember < K) || (noOfSen > (int) (alpha*noOfMember)))
		{
			returnValue = 0;
			break;
		}
	}

	return returnValue;
}

int isSatisfyTCloseness_fromEqTuple(double threshold, data *aData, int *globalDist, equivalenceTuple *oneEqTuple, int noOfSenValue, generalizedAllTable *genAllTable, int noOfDim, int noOfTuple)
{
	int returnValue;

	int *localDist;

	double aValue;

	localDist = tclose_getDistFromEqTuple(oneEqTuple, aData, noOfSenValue);

	aValue = tclose_getSimilarValue(genAllTable, noOfDim, noOfTuple, noOfSenValue, globalDist, localDist, noOfTuple, oneEqTuple->noOfMember);


	if (aValue > threshold)
	{
		returnValue = 0;
	}
	else
	{
		returnValue = 1;
	}

	free(localDist);

	return returnValue;
}

int isSatisfyTCloseness(double threshold, data *aData, int *globalDist, equivalenceClass *eqClass, configure *conf, int noOfSenValue, int noOfDim, int noOfTuple, generalizedAllTable *genAllTable)
{
	int noOfEqTuple;
	equivalenceTuple **eqTuple;
	equivalenceTuple *oneEqTuple;

	int returnValue;

	//equivalenceTuple *oneEqTuple;

	//int noOfMember, noOfSen;

	double alpha;

	int K;

	int i;
	int isSatisfy;

	K = conf->K;

	noOfEqTuple = eqClass->noOfEqClass;
	eqTuple = eqClass->eqClass;

	alpha = conf->alpha;
	
	returnValue = 1;
	//***RAY to be completed *****

	for (i = 0; i < noOfEqTuple; i++)
	{
		oneEqTuple = eqTuple[i];
		isSatisfy = isSatisfyTCloseness_fromEqTuple(threshold, aData, globalDist, oneEqTuple, noOfSenValue, genAllTable, noOfDim, noOfTuple);
		if (isSatisfy == 0)
		{
			returnValue = 0;
			break;
		}
	}

	return returnValue;
}


double *checkAllGenDomainSet(AllGeneralizationDomainSet *allGenDomainSet, equivalenceClass *eqClass, configure *conf, data *dataSet, generalizedAllTable *genAllTable, int isCalculate)
{
	double *returnValue;

	int noOfSet;
	GeneralizationDomainSet *set;

	int canReturn;
	
	equivalenceClass *projEqClass;
	equivalenceClass *cloneProjEqClass;

	AttributeList *attrList;

	int i, j;

	int noOfDim, noOfDomain;

	int isSatisfy;

	double distortion;

	GeneralizationDomain *domain;
	GeneralizationDomain *genDomain;
	GeneralizationDomainSet *genDomainSet;


	noOfSet = allGenDomainSet->noOfSet;
	set = allGenDomainSet->set;

	noOfDim = conf->noOfDim;

	if (isCalculate == 1)
	{
		if (noOfSet == 1)
		{
			canReturn = 1;
		}
		else
		{
			canReturn = 0;
		}
	}
	else
	{
		canReturn = 0;
	}


	for (i = 0; i < noOfSet; i++)
	{ 
		genDomainSet = &(set[i]);


		// create the projected eq. classes
		attrList = &(genDomainSet->attrList);
		projEqClass = cloneEquivalenceClassWithProjection(eqClass, noOfDim, attrList, dataSet, conf);

		//fprintf(g_debugFilePtr, "############## eqClass ##################\n");
		//printf("############## eqClass ##################\n");
		//printEquivalenceClass(g_debugFilePtr, eqClass, conf->noOfDim);

		//fprintf(g_debugFilePtr, "############## projEqClass ##################\n");
		//printf("############## projEqClass ##################\n");
		//printEquivalenceClass(g_debugFilePtr, projEqClass, conf->noOfDim);

		domain = genDomainSet->domain;
		noOfDomain = genDomainSet->noOfDomain;

		// returnValue
		if (canReturn == 1)
		{
			returnValue = (double *) malloc(sizeof(double)*noOfDomain);
			for (j = 0; j < noOfDomain ; j++)
			{
				returnValue[j] = -1;
			}
		}

		for (j = 0; j < noOfDomain; j++)
		{
			genDomain = &(domain[j]);

			if (genDomain->isAlpha == 0)
			{
				// clone
				cloneProjEqClass = cloneEquivalenceClass(projEqClass, conf);

				//fprintf(g_debugFilePtr, "############## cloneProjEqClass (before) ##################\n");
				//printf("############## cloneProjEqClass (before) ##################\n");
				//printEquivalenceClass(g_debugFilePtr, cloneProjEqClass, conf->noOfDim);

				// generalize
				fullSpaceGeneralize(cloneProjEqClass, dataSet, conf, genAllTable, attrList, genDomain);

				//fprintf(g_debugFilePtr, "############## cloneProjEqClass (after) ##################\n");
				//printf("############## cloneProjEqClass (after) ##################\n");
				//printEquivalenceClass(g_debugFilePtr, cloneProjEqClass, conf->noOfDim);

				// check isAlpha
				isSatisfy = isSatisfyKAnonymity(cloneProjEqClass, conf);
				if (isSatisfy == 1)
				{
					//
					genDomain->isAlpha = 1;
					markAllParent(genDomainSet, j);

					if (isCalculate == 1)
					{
						// calculate the distortion
						distortion = calculateDistortion(cloneProjEqClass, conf, genAllTable);

						returnValue[j] = distortion;
					}
				}

				// free
				freeEquivalenceClass(cloneProjEqClass);
				free(cloneProjEqClass);
			}
			else
			{
				// do nothing
			}
		}

		freeEquivalenceClass(projEqClass);
		free(projEqClass);
	}

	if (canReturn == 0)
	{
		returnValue = NULL;
	}

	return returnValue;

}



double *checkAllGenDomainSet_TCloseness(double threshold, int *globalDist, AllGeneralizationDomainSet *allGenDomainSet, equivalenceClass *eqClass, configure *conf, data *dataSet, generalizedAllTable *genAllTable, int isCalculate)
{
	double *returnValue;

	int noOfSet;
	GeneralizationDomainSet *set;

	int canReturn;
	
	equivalenceClass *projEqClass;
	equivalenceClass *cloneProjEqClass;

	AttributeList *attrList;

	int i, j;

	int noOfDim, noOfDomain;

	int isSatisfy;

	double distortion;
	int noOfTuple;
	int noOfSenValue;

	GeneralizationDomain *domain;
	GeneralizationDomain *genDomain;
	GeneralizationDomainSet *genDomainSet;


	noOfSet = allGenDomainSet->noOfSet;
	set = allGenDomainSet->set;

	noOfDim = conf->noOfDim;
	noOfTuple = conf->noOfTuple;
	noOfSenValue = dataSet->dimSize[noOfDim-1];

	if (isCalculate == 1)
	{
		if (noOfSet == 1)
		{
			canReturn = 1;
		}
		else
		{
			canReturn = 0;
		}
	}
	else
	{
		canReturn = 0;
	}


	for (i = 0; i < noOfSet; i++)
	{ 
		genDomainSet = &(set[i]);


		// create the projected eq. classes
		attrList = &(genDomainSet->attrList);
		projEqClass = cloneEquivalenceClassWithProjection(eqClass, noOfDim, attrList, dataSet, conf);

		//fprintf(g_debugFilePtr, "############## eqClass ##################\n");
		//printf("############## eqClass ##################\n");
		//printEquivalenceClass(g_debugFilePtr, eqClass, conf->noOfDim);

		//fprintf(g_debugFilePtr, "############## projEqClass ##################\n");
		//printf("############## projEqClass ##################\n");
		//printEquivalenceClass(g_debugFilePtr, projEqClass, conf->noOfDim);

		domain = genDomainSet->domain;
		noOfDomain = genDomainSet->noOfDomain;

		// returnValue
		if (canReturn == 1)
		{
			returnValue = (double *) malloc(sizeof(double)*noOfDomain);
			for (j = 0; j < noOfDomain ; j++)
			{
				returnValue[j] = -1;
			}
		}

		for (j = 0; j < noOfDomain; j++)
		{
			genDomain = &(domain[j]);

			if (genDomain->isAlpha == 0)
			{
				// clone
				cloneProjEqClass = cloneEquivalenceClass(projEqClass, conf);

				//fprintf(g_debugFilePtr, "############## cloneProjEqClass (before) ##################\n");
				//printf("############## cloneProjEqClass (before) ##################\n");
				//printEquivalenceClass(g_debugFilePtr, cloneProjEqClass, conf->noOfDim);

				// generalize
				fullSpaceGeneralize(cloneProjEqClass, dataSet, conf, genAllTable, attrList, genDomain);

				//fprintf(g_debugFilePtr, "############## cloneProjEqClass (after) ##################\n");
				//printf("############## cloneProjEqClass (after) ##################\n");
				//printEquivalenceClass(g_debugFilePtr, cloneProjEqClass, conf->noOfDim);

				// check isAlpha
				isSatisfy = isSatisfyTCloseness(threshold, dataSet, globalDist, cloneProjEqClass, conf, noOfSenValue, noOfDim, noOfTuple, genAllTable);
				if (isSatisfy == 1)
				{
					//
					genDomain->isAlpha = 1;
					markAllParent(genDomainSet, j);

					if (isCalculate == 1)
					{
						// calculate the distortion
						distortion = calculateDistortion(cloneProjEqClass, conf, genAllTable);

						returnValue[j] = distortion;
					}
				}

				// free
				freeEquivalenceClass(cloneProjEqClass);
				free(cloneProjEqClass);
			}
			else
			{
				// do nothing
			}
		}

		freeEquivalenceClass(projEqClass);
		free(projEqClass);
	}

	if (canReturn == 0)
	{
		returnValue = NULL;
	}

	return returnValue;

}



// no. of domain <= 0
int chooseBestGenDomain(GeneralizationDomainSet *genDomainSet, double *dist)
{
	int returnValue;
	int noOfDomain;
	GeneralizationDomain *domain;

	double minDist; 

	int i;

	returnValue = -1;

	domain = genDomainSet->domain;
	noOfDomain = genDomainSet->noOfDomain;

	if (noOfDomain > 0)
	{
		//minDist = dist[0];
		//returnValue = -1;
		minDist = -1;
		for (i = 0; i < noOfDomain; i++)
		{
			if ((dist[i] != -1) && ((dist[i] < minDist) || (minDist == -1)))
			{
				minDist = dist[i];
				returnValue = i;
			}
		}
	}
	else
	{
		returnValue = -1;
	}

	return returnValue;
	
}

equivalenceClass *chooseBestFinalEqClass(AllGeneralizationDomainSet *allGenDomainSet, double *dist, equivalenceClass *eqClass, data *dataSet, configure *conf, generalizedAllTable *genAllTable)
{
	equivalenceClass *returnValue;

	GeneralizationDomainSet *genDomainSet;

	GeneralizationDomain *genDomain;
	AttributeList *attrList;

	int noOfSet;
	GeneralizationDomainSet *set;

	int bestIndex;

	noOfSet = allGenDomainSet->noOfSet;
	set = allGenDomainSet->set;

	if (noOfSet == 1)
	{
		genDomainSet = &(set[0]);

		// find best
		bestIndex = chooseBestGenDomain(genDomainSet, dist);

		if (bestIndex == -1)
		{
			returnValue = NULL;
		}
		else
		{
			// clone
			returnValue = cloneEquivalenceClass(eqClass, conf);

			// generalize
			attrList = &(genDomainSet->attrList);
			genDomain = &(genDomainSet->domain[bestIndex]);
			fullSpaceGeneralize(returnValue, dataSet, conf, genAllTable, attrList, genDomain);
		}
	}
	else
	{
		returnValue = NULL;
	}

	
	return returnValue;
}


void tclose_singleExecution(data *aData, configure *config, generalizedAllTable *genAllTable)
{
	equivalenceClass *proEqClass;

	int noOfSenValue;
	int noOfDim;

	int m;

	int noOfViolate;

	tuple *tupleList;

	signatureGroupGroupGroup *allSig;

	int ***allSigMapping;

	FILE *avgRelError_filePtr;
	
	allCondition *aAllCondition;
	int noOfQID;

	twoTable *finalTable;

	double avgRelError;

	double avgEqTupleSize, avgSenEqTupleSize;

	mechAllGroup *finalMechAllGroup;

	FILE *debugFilePtr;
	char *debugFilename = "debug.txt";
	int senIndex;

	//int dummy;
	double infoLoss, distortion;
	int noOfSenChange;

	double globalDistDiff;
	double globalDistDiffSen;

	int noOfSenIndex;

	int threshold;

	int noOfEqTupleSenValue;

	char *avgRelErrorFilename = "error.txt";
	char *timeFilename = "time.txt";

	int queryNoOfDim = config->queryNoOfDim;
	double s = config->s;
	int noOfCondition = config->noOfCondition;

#ifndef _WIN32
	FILE *timeFilePtr;

	float  userTime, sysTime;
	float userTime_dist, sysTime_dist;
	struct rusage myTime1, myTime2, myTime_distEnd;

	getrusage(RUSAGE_SELF,&myTime1);
#endif

		noOfDim = aData->noOfDim;
	noOfSenValue = aData->dimSize[noOfDim-1];
	noOfQID = noOfDim-1;
	m = config->d;

	tupleList = aData->tupleList;


	// signature
	threshold = calViolation_calculateSampleSize(config->error, config->sigma);
	//threshold = 1;
	printf(" Sample threshold = %d\n", threshold);
	

	// assume only one sensitive index
	noOfSenIndex = 2;



	// generate all signature
	printf("Generating signatures...\n");
	allSig = sig_findAllSignature(aData, threshold);

	sig_printStatistics(stdout, allSig);

	// find distribution
	printf("Finding the global distributions...\n");
	allSigMapping = sig_findGlobalDistributionFromOrgTable(config, allSig, aData, noOfSenIndex);

#ifndef _WIN32

	getrusage(RUSAGE_SELF, &myTime_distEnd);	

	userTime =
     		((float) (myTime_distEnd.ru_utime.tv_sec  - myTime1.ru_utime.tv_sec)) +
       		((float) (myTime_distEnd.ru_utime.tv_usec - myTime1.ru_utime.tv_usec)) * 1e-6;
	sysTime =
       		((float) (myTime_distEnd.ru_stime.tv_sec  - myTime1.ru_stime.tv_sec)) +
       		((float) (myTime_distEnd.ru_stime.tv_usec - myTime1.ru_stime.tv_usec)) * 1e-6;


	userTime_dist = userTime;
	sysTime_dist = sysTime;
#endif

	

	//
	debugFilePtr = (FILE *) fopen(debugFilename, "w");
	if (debugFilePtr == NULL)
	{
		printf("File %s cannot be opened!\n", debugFilename);
		exit(0);
	}

	proEqClass = fullSpace_TCloseness(aData, config, genAllTable, debugFilePtr);

	//
	//printf("Final equivalence class...\n");
	//printEquivalenceClass(debugFilePtr, proEqClass, noOfDim);

	//
	equivalenceClass_generateGeneralizedValue(proEqClass, aData, genAllTable, noOfDim);
	createSenValueInEqClass(proEqClass, aData, noOfSenValue);
	eq_createSenIndexForAllEqTuple(proEqClass, config, noOfSenValue, noOfSenIndex);


	// transform
	printf("Transform an equivalence class to mechGroup...\n");
	createSenValueInEqClass(proEqClass, aData, noOfSenValue);
	copySenValueToNewSenValueInEqClass(proEqClass, noOfSenValue);
	finalMechAllGroup = transformFromEqClassToMechAllGroup(proEqClass, noOfSenValue);
#ifndef _WIN32
	getrusage(RUSAGE_SELF, &myTime2);	
#endif
	//
	//allCondition *predicate_generateAllCondition(int noOfCondition, int queryNoOfDim, int noOfQID, double s, int *dimSize)
	printf("Generating predicates...\n");
	aAllCondition = predicate_generateAllCondition(noOfCondition, queryNoOfDim, noOfQID, s, aData->dimSize);

	// generate two tables
	printf("Generating two tables...\n");
	finalTable = generateTwoTableFromMechGroup(finalMechAllGroup, aData);

	// printing
	//printTwoTable(debugFilePtr, finalTable);

	// calculate the estimated count and actual count
	avgRelError = calculateAvgRelativeError(finalMechAllGroup, aAllCondition, aData, noOfQID);

	printf("Average relative error : %.2f\n", avgRelError);

	

	// calculate the coverage error (information loss and distortion)
	//void calculateQIDInfoLossAndSenDistortion(int noOfQID, int totalNoOfTuple, equivalenceClass *eqClass, generalizedAllTable *genAllTable, double *infoLoss, double *distortion)
	calculateQIDInfoLossAndSenDistortion(noOfQID, config->noOfTuple, proEqClass, genAllTable, &infoLoss, &distortion, &noOfSenChange);


	printf("Information Loss : %.2f\n", infoLoss);
	printf("Distortion       : %.2f\n", distortion);
	printf("No. of Sen Change: %d\n", noOfSenChange);


	avgRelError_filePtr = (FILE *) fopen(avgRelErrorFilename, "w");
	fprintf(avgRelError_filePtr, "%.6f\n", avgRelError);
	fprintf(avgRelError_filePtr, "%.6f\n", infoLoss);
	fprintf(avgRelError_filePtr, "%.6f\n", distortion);
	fprintf(avgRelError_filePtr, "%d\n", noOfSenChange);
	

	// assume one senIndex
	printf("Transforming ...\n");
	transformToBinaryData(config, proEqClass, aData);

	senIndex = 1;
	avgEqTupleSize = ass_findAvgEqTupleSize(proEqClass);
	avgSenEqTupleSize = ass_findAvgSenEqTupleSize(proEqClass, senIndex);
	fprintf(avgRelError_filePtr, "%.6f\n", avgEqTupleSize);
	fprintf(avgRelError_filePtr, "%.6f\n", avgSenEqTupleSize);

	globalDistDiff = ass_findAvgGlobalDistDiff(proEqClass, senIndex, allSig, allSigMapping, m);
	globalDistDiffSen = ass_findAvgGlobalDistDiffSen(proEqClass, senIndex, allSig, allSigMapping, m);
	fprintf(avgRelError_filePtr, "%.10f\n", globalDistDiff);
	fprintf(avgRelError_filePtr, "%.10f\n", globalDistDiffSen);

	//
	noOfEqTupleSenValue = anatomy_findNoOfEqTupleWithSenIndex(proEqClass, senIndex);
	noOfViolate = anatomy_findNoOfViolateEqTuple(proEqClass, senIndex, allSig, allSigMapping, m, aData->noOfTuple);
	fprintf(avgRelError_filePtr, "\n");
	fprintf(avgRelError_filePtr, "%d\n", proEqClass->noOfEqClass);	// total. no. of eq tuples
	fprintf(avgRelError_filePtr, "%d\n", noOfEqTupleSenValue);			// total no. of eq tuples with sensitive values
	fprintf(avgRelError_filePtr, "%d\n", noOfViolate);				// total no. of eq. tuples with privacy breach


	printf("\n");
	printf("Avg. eq. tuple size = %.6f\n", avgEqTupleSize);
	printf("Avg. sen. eq. tuple size = %.6f\n", avgSenEqTupleSize);
	printf("Global dist. diff = %.10f\n", globalDistDiff);
	printf("Global dist. diff (sen) = %.10f\n", globalDistDiffSen);
	printf("Total no. of eq. classes = %d\n", proEqClass->noOfEqClass);
	printf("Total no. of eq. classes with sen. values = %d\n", noOfEqTupleSenValue);
	printf("Total no. of eq. classes with privacy breach = %d\n", noOfViolate);
	printf("\n");


	fclose(avgRelError_filePtr);


#ifndef _WIN32
	userTime =
     		((float) (myTime2.ru_utime.tv_sec  - myTime1.ru_utime.tv_sec)) +
       		((float) (myTime2.ru_utime.tv_usec - myTime1.ru_utime.tv_usec)) * 1e-6;
	sysTime =
       		((float) (myTime2.ru_stime.tv_sec  - myTime1.ru_stime.tv_sec)) +
       		((float) (myTime2.ru_stime.tv_usec - myTime1.ru_stime.tv_usec)) * 1e-6;

	printf("User time : %f seconds\n",userTime);
	printf("System time : %f seconds\n\n",sysTime);

	fprintf(debugFilePtr, "Execution Time\n");
	fprintf(debugFilePtr, "--------------\n");
	fprintf(debugFilePtr, "User time : %f seconds\n",userTime);
    fprintf(debugFilePtr, "System time : %f seconds\n\n",sysTime);


	timeFilePtr = (FILE *) fopen(timeFilename, "w");
	fprintf(timeFilePtr, "%.6f\n",userTime);
    fprintf(timeFilePtr, "%.6f\n\n",sysTime);
	fprintf(timeFilePtr, "%.6f\n",userTime_dist);
    fprintf(timeFilePtr, "%.6f\n\n",sysTime_dist);
	fprintf(timeFilePtr, "%.6f\n",userTime-userTime_dist);
    fprintf(timeFilePtr, "%.6f\n\n",sysTime-sysTime_dist);
	fclose(timeFilePtr);
#endif

	fclose(debugFilePtr);
}


int main()
{
	char *configFilename = "conf.txt";

	configure *config;
	data *dataSet;
	generalizedAllTable *genAllTable;

	config = readConfig(configFilename);
	dataSet = readData(config);
	genAllTable = readGen(config);

	printf("Calculating the coverage...\n");
	//void calculateCoverage(generalizedAllTable *genAllTable, int noOfDim)
	calculateCoverage(genAllTable, config->noOfDim);

	//void progressiveMain(data *dataSet, configure *conf, generalizedAllTable *genAllTable)
	//progressiveMain(dataSet, config, genAllTable);

	//hierachicalMain(dataSet, config, genAllTable);
	//randomMain(dataSet, config, genAllTable);
	//fullSpaceMain(dataSet, config, genAllTable);

	//topDownAlphaMain(dataSet, config, genAllTable);

	//mechMain(dataSet, config, genAllTable);
	//mechDistortionMain(dataSet, config, genAllTable);

	//calCredMain(dataSet, config, genAllTable);

	if (config->alg == 30)
	{
		calViolationMain(dataSet, config, genAllTable);
	}
	else if (config->alg == 31)
	{
		// do sth
		ass_main(config, dataSet, genAllTable);
	}
	else if (config->alg == 41)
	{
		anatomy_main(dataSet, config, genAllTable);
	}
	else if (config->alg == 42)
	{
		mechDistortionMain(dataSet, config, genAllTable);
	}
	else if (config->alg == 43)
	{
		injector_singleExecution(dataSet, config->d, config, genAllTable);
	}
	else if (config->alg == 44)
	{
		tclose_singleExecution(dataSet, config, genAllTable);
	}
	else if (config->alg == 51)
	{
		// injector violation (implicity with algNo = 51)
		calViolationMain(dataSet, config, genAllTable);
	}
	else if (config->alg == 52)
	{
		// t-closeness violation
		calViolationMain(dataSet, config, genAllTable);
	}
	else if (config->alg == 53)
	{
		// MASK violation
		calViolationMain(dataSet, config, genAllTable);
	}



	//testQsort();

	return 0;
}
