#include "set.h"

mySet *mySet_new()
{
	mySet *returnValue;

	returnValue = (mySet *) malloc(sizeof(mySet));
	memset(returnValue, 0, sizeof(mySet));

	return returnValue;
}


void mySet_insert(mySet *aSet, int eltNo)
{
	int noOfElt;

	noOfElt = aSet->noOfElt;
	if (noOfElt == 0)
	{
		aSet->elt = (int *) malloc(sizeof(int));
	}
	else
	{
		aSet->elt = (int *) realloc(aSet->elt, sizeof(int)*(noOfElt+1));
	}
	aSet->elt[noOfElt] = eltNo;
	aSet->noOfElt++;
}

void mySet_free(mySet *aSet)
{
	if (aSet->noOfElt > 0)
	{
		free(aSet->elt);
	}
	free(aSet);
}

// 
int mySet_isExist(mySet *aSet, int aValue)
{
	int returnValue;

	int noOfElt;
	int eltNo;
	int *elt;

	int i;

	noOfElt = aSet->noOfElt;
	elt = aSet->elt;

	returnValue = 0;
	for (i = 0; i < noOfElt; i++)
	{
		eltNo = elt[i];

		if (eltNo == aValue)
		{
			returnValue = 1;
			break;
		}
	}

	return returnValue;
}

void mySet_insertNonRepeat(mySet *aSet, int aValue)
{
	int isExist;

	isExist = mySet_isExist(aSet, aValue);
	if (isExist == 0)
	{
		mySet_insert(aSet, aValue);
	}
}

void mySet_unionNonRepeat(mySet *set1, mySet *set2)
{
	int noOfElt2;
	int *elt2;

	int aValue;
	int i;

	elt2 = set2->elt;
	noOfElt2 = set2->noOfElt;

	for (i = 0; i < noOfElt2; i++)
	{
		aValue = elt2[i];
		mySet_insertNonRepeat(set1, aValue);
	}
}

int mySet_isOverlap(mySet *set1, mySet *set2)
{
	int returnValue;

	int noOfElt1;
	int *elt1;

	int noOfElt2;
	int *elt2;

	int i, j;

	elt1 = set1->elt;
	noOfElt1 = set1->noOfElt;

	elt2 = set2->elt;
	noOfElt2 = set2->noOfElt;

	returnValue = 0;
	for (i = 0; i < noOfElt1 ; i++)
	{
		for (j = 0; j < noOfElt2; j++)
		{
			if (elt1[i] == elt2[j])
			{
				returnValue = 1;
				break;
			}
		}

		if (returnValue == 1)
		{
			break;
		}			
	}

	return returnValue;
}

// assume sorted with these two sets
int mySet_isEqual(mySet *aSet1, mySet *aSet2)
{
	int returnValue;

	int noOfElt1, noOfElt2;
	int *elt1;
	int *elt2;

	int i;

	noOfElt1 = aSet1->noOfElt;
	elt1 = aSet1->elt;

	noOfElt2 = aSet2->noOfElt;
	elt2 = aSet2->elt;

	if (noOfElt1 != noOfElt2)
	{
		returnValue = 0;
	}
	else
	{
		returnValue = 1;
		for (i = 0; i < noOfElt1; i++)
		{
			if (elt1[i] != elt2[i])
			{
				returnValue = 0;
				break;
			}
		}
	}
	
	return returnValue;
}

mySet *mySet_clone(mySet *aSet)
{
	mySet *returnValue;

	returnValue = mySet_new();

	returnValue->noOfElt = aSet->noOfElt;
	returnValue->elt = (int *) malloc(sizeof(int)*returnValue->noOfElt);
	memcpy(returnValue->elt, aSet->elt, sizeof(int)*returnValue->noOfElt);

	return returnValue;
}

void mySet_removeIndex(mySet *aSet, int index)
{
	int *elt;
	int noOfElt;

	int *newElt;

	elt = aSet->elt;
	noOfElt = aSet->noOfElt;

	if ((index < 0) || (index >= noOfElt))
	{
		return;
	}

	if (noOfElt == 1)
	{
		// free
		free(elt);

		aSet->noOfElt = 0;
		aSet->elt = NULL;
	}
	else
	{
		newElt = (int *) malloc(sizeof(int)*(noOfElt-1));

		if (index != 0)
		{
			memcpy(newElt, elt, sizeof(int)*index);
		}

		if (index != noOfElt-1)
		{
			memcpy(&(newElt[index]), &(elt[index+1]), sizeof(int)*(noOfElt-index-1));
		}

		aSet->elt = newElt;
		aSet->noOfElt--;

		free(elt);
	}
}
