
#include "signature.h"


signature *sig_new()
{
	signature *returnValue;

	returnValue = (signature *) malloc(sizeof(signature));
	memset(returnValue, 0, sizeof(signature));

	return returnValue;
}

void sig_insertMember(signature *aSignature, int memberID)
{
	int noOfElt;
	int *elt;

	noOfElt = aSignature->noOfMember;
	if (noOfElt == 0)
	{
		elt = (int *) malloc(sizeof(int));
	}
	else
	{
		elt = (int *) realloc(aSignature->member, sizeof(int)*(noOfElt+1));
		
	}
	elt[noOfElt] = memberID;

	aSignature->member = elt;
	aSignature->noOfMember++;
}

void sig_insertContent(signature *aSignature, int dimNo, int level, int value)
{
	int noOfAttr;
	int *dimNoList;
	dimension *content;

	noOfAttr = aSignature->noOfAttr;
	if (noOfAttr == 0)
	{
		dimNoList = (int *) malloc(sizeof(int));
		content = (dimension *) malloc(sizeof(dimension));
	}
	else
	{
		dimNoList = (int *) realloc(aSignature->dimNo, sizeof(int)*(noOfAttr+1));
		content = (dimension *) realloc(aSignature->content, sizeof(dimension)*(noOfAttr+1));
	}
	content[noOfAttr].level = level;
	content[noOfAttr].value = value;
	dimNoList[noOfAttr] = dimNo;

	aSignature->content = content;
	aSignature->dimNo = dimNoList;
	aSignature->noOfAttr++;

}


signatureGroup *sigGroup_new()
{
	signatureGroup *returnValue;

	returnValue = (signatureGroup *) malloc(sizeof(signatureGroup));
	memset(returnValue, 0, sizeof(signatureGroup));

	return returnValue;
}

void sigGroup_insertSig(signatureGroup *aSignatureGroup, signature *aSignature)
{
	int noOfSig;

	signature **sigList;

	noOfSig = aSignatureGroup->noOfSig;
	if (noOfSig == 0)
	{
		sigList = (signature **) malloc(sizeof(signature *));
	}
	else
	{
		sigList = (signature **) realloc(aSignatureGroup->sigList, sizeof(signature *)*(noOfSig+1));
	}
	sigList[noOfSig] = aSignature;

	aSignatureGroup->noOfSig++;
	aSignatureGroup->sigList = sigList;
}


// assume the published table has the same tuples as the origianl table
signatureGroup *sig_createRawSignatureForSingleAttr(int dimNo, data *aData)
{
	signatureGroup *returnValue;

	int dimSize;
	signature *aSignature;

	int i;

	tuple *tupleList;

	signature **sigList;

	int level, value;

	int memberID;

	returnValue = sigGroup_new();

	dimSize = aData->dimSize[dimNo];

	returnValue->noOfSig = dimSize;
	sigList = (signature **) malloc(sizeof(signature *)*returnValue->noOfSig);
	returnValue->sigList = sigList;

	for (i = 0; i < returnValue->noOfSig; i++)
	{
		aSignature = sig_new();
		sigList[i] = aSignature;

		level = 0;
		value = i;
		sig_insertContent(aSignature, dimNo, level, value);
	}

	// scan
	tupleList = aData->tupleList;
	for (i = 0; i < aData->noOfTuple; i++)
	{
		// use value
		value = tupleList[i].content[dimNo].value;
		
		aSignature = sigList[value];
		memberID = i;
		sig_insertMember(aSignature, memberID);
	}

	return returnValue;
}

void sig_free(signature *aSignature)
{
	int noOfAttr, noOfMember;

	if (aSignature != NULL)
	{
		noOfAttr = aSignature->noOfAttr;
		noOfMember = aSignature->noOfMember;

		if (noOfAttr > 0)
		{
			free(aSignature->content);
			free(aSignature->dimNo);
		}

		if (noOfMember > 0)
		{
			free(aSignature->member);
		}
	}
}


signatureGroup *sig_createRealSignatureForSingleAttr(int dimNo, data *aData, int threshold)
{
	signatureGroup *returnValue;

	signatureGroup *tempGroup;
	signature *aSignature;
	int i;

	//int *isFlag;


	returnValue = sigGroup_new();


	tempGroup = sig_createRawSignatureForSingleAttr(dimNo, aData);


	//isFlag = (int *) malloc(sizeof(int)*tempGroup->noOfSig);
	//memset(isFlag, 0, sizeof(int)*tempGroup->noOfSig);


	for (i = 0; i < tempGroup->noOfSig; i++)
	{
		aSignature = tempGroup->sigList[i];

		if (aSignature->noOfMember >= threshold)
		{
			sigGroup_insertSig(returnValue, aSignature);
		}
		else
		{
			sig_free(aSignature);
		}
	}
	//free(isFlag);
	if (tempGroup->noOfSig > 0)
	{
		free(tempGroup->sigList);
	}
	free(tempGroup);

	return returnValue;
}


signatureGroupGroup *sigGroupGroup_new()
{
	signatureGroupGroup *returnValue;

	returnValue = (signatureGroupGroup *) malloc(sizeof(signatureGroupGroup));
	memset(returnValue, 0, sizeof(signatureGroupGroup));

	return returnValue;
}

void sigGroupGroup_insert(signatureGroupGroup *aSigGroupGroup, signatureGroup *aSigGroup)
{
	int noOfElt;
	signatureGroup **elt;
	signatureGroup **newElt;

	elt = aSigGroupGroup->groupList;
	noOfElt = aSigGroupGroup->noOfGroup;

	if (noOfElt == 0)
	{
		newElt = (signatureGroup **) malloc(sizeof(signatureGroup *));
	}
	else
	{
		newElt = (signatureGroup **) realloc(elt, sizeof(signatureGroup *)*(noOfElt+1));
	}
	newElt[noOfElt] = aSigGroup;
	//if (noOfElt > 0)
	//{
	//	memcpy(newElt, elt, sizeof(signatureGroup *)*noOfElt);
	//}

	aSigGroupGroup->noOfGroup++;
	aSigGroupGroup->groupList = newElt;
}

signatureGroupGroup *sig_createRealSignatureForMultipleAttr(data *aData, int threshold)
{
	int dimNo;
	int i;

	int noOfQIDAttr;

	signatureGroupGroup *returnValue;

	noOfQIDAttr = aData->noOfDim - 1;

	returnValue = sigGroupGroup_new();
	returnValue->noOfGroup = noOfQIDAttr;
	returnValue->groupList = (signatureGroup **) malloc(sizeof(signatureGroup *)*noOfQIDAttr);

	for (i = 0; i < noOfQIDAttr; i++)
	{
		dimNo = i;

		returnValue->groupList[i] = sig_createRealSignatureForSingleAttr(dimNo, aData, threshold);
	}

	return returnValue;
}

int *sig_createTupleMapSignature(signatureGroup *aSignatureGroup, int noOfTuple)
{
	int *returnValue;
	int i, j;
	int memberID;

	signature **sigList;

	signature *aSignature;

	returnValue = (int *) malloc(sizeof(int)*noOfTuple);
	for (i = 0; i < noOfTuple; i++)
	{
		returnValue[i] = -1;
	}
//printf(" @@@@@@@@@@@@@3.1\n");
	sigList = aSignatureGroup->sigList;
//printf(" @@@@@@@@@@@@@3.2 noOfSig=%d\n", aSignatureGroup->noOfSig);
	for (i = 0; i < aSignatureGroup->noOfSig; i++)
	{
//printf(" @@@@@@@@@@@@@3.3\n");
		aSignature = sigList[i];
//printf(" @@@@@@@@@@@@@3.4, noOfMember=%d\n", aSignature->noOfMember);
		for (j = 0; j < aSignature->noOfMember; j++)
		{
//printf(" @@@@@@@@@@@@@3.5  j=%d\n", j);
			memberID = aSignature->member[j];
//printf("     memberID=%d,noOfTuple=%d,noOfMember=%d,i=%d,j=%d\n", memberID, noOfTuple, aSignature->noOfMember,i,j);
			returnValue[memberID] = i;
		}
	}

	return returnValue;
}



void sig_createSigNoForOneEqClass(equivalenceTuple *oneEqTuple, int *mapTupleList)
{
	int i;
	
	int noOfMember;
	int *member;
	int memberID;
	int sigID;

	int *signatureID;

	noOfMember = oneEqTuple->noOfMember;

	member = oneEqTuple->member;

	signatureID = oneEqTuple->signatureID;
	if (signatureID == NULL)
	{
		signatureID = (int *) malloc(sizeof(int)*noOfMember);
		oneEqTuple->signatureID = signatureID;
	}
	memset(signatureID, 0, sizeof(int)*noOfMember);

	for (i = 0; i < noOfMember; i++)
	{
		memberID = member[i];

		sigID = mapTupleList[memberID];

		signatureID[i] = sigID;
	}
}

void sig_createSigNoForAllEqClass(equivalenceClass *eqClass, int *mapTupleList)
{
	int noOfTuple;
	equivalenceTuple **tupleList;
	equivalenceTuple *oneTuple;

	int i;

	noOfTuple = eqClass->noOfEqClass;
	tupleList = eqClass->eqClass;

	for (i = 0; i < noOfTuple; i++)
	{
		oneTuple = tupleList[i];

		sig_createSigNoForOneEqClass(oneTuple, mapTupleList);
	}
}

int sig_isContainMatch(equivalenceTuple *aEqTuple, int sigID)
{
	int returnValue;

	int i;

	int noOfMember;
	int *signatureID;

	noOfMember = aEqTuple->noOfMember;
	signatureID = aEqTuple->signatureID;

	returnValue = 0;
	for (i = 0; i < noOfMember; i++)
	{
		if (signatureID[i] == sigID)
		{
			returnValue = 1;
			break;
		}
	}

	return returnValue;
}

int *sig_findEqClassWithSignature(equivalenceClass *eqClass, int sigID, int *noOfMatch)
{
	int *returnValue;
	equivalenceTuple **eqClassList;
	equivalenceTuple *aEqTuple;
	int noOfEqTuple;
	int i;

	int *isMatch;

	int index;

	(*noOfMatch) = 0;

	

	noOfEqTuple = eqClass->noOfEqClass;
	eqClassList = eqClass->eqClass;

	isMatch = (int *) malloc(sizeof(int)*noOfEqTuple);
	memset(isMatch, 0, sizeof(int)*noOfEqTuple);

	for (i = 0; i < noOfEqTuple; i++)
	{
		aEqTuple = eqClassList[i];
		isMatch[i] = sig_isContainMatch(aEqTuple, sigID);
		if (isMatch[i] == 1)
		{
			(*noOfMatch)++;
		}
	}

	returnValue = (int *) malloc(sizeof(int)*(*noOfMatch));
	memset(returnValue, 0, sizeof(int)*(*noOfMatch));

	index = 0;
	for (i = 0; i < noOfEqTuple; i++)
	{
		if (isMatch[i] == 1)
		{
			returnValue[index] = i;
			index++;
		}
	}

	free(isMatch);

	return returnValue;
}


// assume dimension is the same
int sig_isSamePrefixValue(signature *aSig1, signature *aSig2)
{
	int returnValue;

	dimension *content1;
	dimension *content2;

	int noOfAttr1, noOfAttr2;

	int i;

	content1 = aSig1->content;
	content2 = aSig2->content;

	noOfAttr1 = aSig1->noOfAttr;
	noOfAttr2 = aSig2->noOfAttr;

	returnValue = 1;
	for (i = 0; i < noOfAttr1-1; i++)
	{
		if (content1[i].value != content1[i].value)
		{
			returnValue = 0;
			break;
		}
	}

	return returnValue;
}

int sig_isSamePrefixDim(signatureGroup *aSigGroup1, signatureGroup *aSigGroup2)
{
	int returnValue;

	int noOfSig1, noOfSig2;
	int noOfAttr1, noOfAttr2;

	int i;

	signature **sigList1;
	signature **sigList2;

	signature *aSig1;
	signature *aSig2;

	noOfSig1 = aSigGroup1->noOfSig;
	noOfSig2 = aSigGroup2->noOfSig;

	sigList1 = aSigGroup1->sigList;
	sigList2 = aSigGroup2->sigList;

	

	if ((noOfSig1 > 0) && (noOfSig2 > 0))
	{
		aSig1 = sigList1[0];
		aSig2 = sigList2[0];

		noOfAttr1 = aSig1->noOfAttr;
		noOfAttr2 = aSig2->noOfAttr;

		returnValue = 1;
		for (i = 0; i < noOfAttr1-1; i++)
		{
			if (aSig1->dimNo[i] != aSig2->dimNo[i])
			{
				returnValue = 0;
				break;
			}
		}
	}
	else
	{
		returnValue = 0;
	}

	return returnValue;
}

// assume the member is sorted
int *sig_combineMember(signature *aSig1, signature *aSig2, int *noOfFinalMember)
{
	int *returnValue;

	int i, j;

	int *isFlag;

	int noOfMember1, noOfMember2;
	int *member1;
	int *member2;

	int index;

	int isExist;

	noOfMember1 = aSig1->noOfMember;
	noOfMember2 = aSig2->noOfMember;

	member1 = aSig1->member;
	member2 = aSig2->member;

	isFlag = (int *) malloc(sizeof(int)*noOfMember1);
	memset(isFlag, 0, sizeof(int)*noOfMember1);

	(*noOfFinalMember) = 0;
	for (i = 0; i < noOfMember1; i++)
	{
		isExist = 0;
		for (j = 0; j < noOfMember2; j++)
		{
			if (member1[i] == member2[j])
			{
				isExist = 1;
				break;
			}

			if (member2[j] > member1[i])
			{
				break;
			}
		}

		if (isExist == 1)
		{
			isFlag[i] = 1;
			(*noOfFinalMember)++;
		}
	}

	returnValue = (int *) malloc(sizeof(int)*(*noOfFinalMember));
	index = 0;
	for (i = 0; i < noOfMember1; i++)
	{
		if (isFlag[i] == 1)
		{
			returnValue[index] = member1[i];
			index++;
		}
	}


	free(isFlag);

	return returnValue;
}


signature *sig_combineTwoSigAndGenerateNewSig(signature *aSig1, signature *aSig2)
{
	signature *returnValue;

	dimension *content1;
	dimension *content2;

	int *dimNo1;
	int *dimNo2;

	int noOfAttr;
	int finalNoOfAttr;

	int i;

	content1 = aSig1->content;
	content2 = aSig2->content;

	dimNo1 = aSig1->dimNo;
	dimNo2 = aSig2->dimNo;

	noOfAttr = aSig1->noOfAttr;

	finalNoOfAttr = noOfAttr+1;

	returnValue = sig_new();

	returnValue->content = (dimension *) malloc(sizeof(dimension)*finalNoOfAttr);
	returnValue->dimNo = (int *) malloc(sizeof(int)*finalNoOfAttr);
	for (i = 0; i < finalNoOfAttr-1; i++)
	{
		returnValue->content[i] = content1[i];
		returnValue->dimNo[i] = dimNo1[i];
	}	
	returnValue->content[finalNoOfAttr-1] = content2[noOfAttr-1];
	returnValue->dimNo[finalNoOfAttr-1] = dimNo2[noOfAttr-1];

	returnValue->noOfAttr = finalNoOfAttr;

	returnValue->member = sig_combineMember(aSig1, aSig2, &(returnValue->noOfMember));

	return returnValue;
}

signatureGroup *signatureRawSignatureForCombineAttrFromTwoSigGroup(signatureGroup *sigGroup1, signatureGroup *sigGroup2) 
{
	signatureGroup *returnValue;

	signature *aSignature;

	int noOfSig1, noOfSig2;
	signature **sigList1;
	signature **sigList2;

	signature *aSig1;
	signature *aSig2;

	int i, j;

	int isSamePrefixDim, isSamePrefixValue;

	
	isSamePrefixDim = sig_isSamePrefixDim(sigGroup1, sigGroup2);
	if (isSamePrefixDim == 1)
	{
		returnValue = sigGroup_new();

		noOfSig1 = sigGroup1->noOfSig;
		noOfSig2 = sigGroup2->noOfSig;

		sigList1 = sigGroup1->sigList;
		sigList2 = sigGroup2->sigList;

		for (i = 0; i < noOfSig1; i++)
		{
			aSig1 = sigList1[i];
			for (j = 0; j < noOfSig2; j++)
			{
				aSig2 = sigList2[j];

				isSamePrefixValue = sig_isSamePrefixValue(aSig1, aSig2);
				if (isSamePrefixValue == 1)
				{
					aSignature = sig_combineTwoSigAndGenerateNewSig(aSig1, aSig2);

					sigGroup_insertSig(returnValue, aSignature);
				}
			}
		}
	}
	else
	{
		returnValue = NULL;
	}

	return returnValue;
}



signatureGroup *signatureRealSignatureForCombineAttrFromTwoSigGroup(signatureGroup *sigGroup1, signatureGroup *sigGroup2, int threshold) 
{
	signatureGroup *returnValue;
	signatureGroup *rawSigGroup;

	int noOfSatisfy;
	int *isFlag;

	signature *aSig;

	int i;
	int index;

	

	rawSigGroup = signatureRawSignatureForCombineAttrFromTwoSigGroup(sigGroup1, sigGroup2);

	if (rawSigGroup == NULL)
	{
		returnValue = NULL;
	}
	else
	{
		returnValue = sigGroup_new();

		isFlag = (int *) malloc(sizeof(int)*rawSigGroup->noOfSig);
		memset(isFlag, 0, sizeof(int)*rawSigGroup->noOfSig);

		noOfSatisfy = 0;
		for (i = 0; i < rawSigGroup->noOfSig; i++)
		{
			aSig = rawSigGroup->sigList[i];
			if (aSig->noOfMember >= threshold)
			{
				isFlag[i] = 1;
				noOfSatisfy++;
			}
		}

		returnValue->noOfSig = noOfSatisfy;
		returnValue->sigList = (signature **) malloc(sizeof(signature *)*noOfSatisfy);
		index = 0;
		for (i = 0; i < rawSigGroup->noOfSig; i++)
		{
			aSig = rawSigGroup->sigList[i];
			if (isFlag[i] == 1)
			{
				returnValue->sigList[index] = aSig;
				index++;
			}
			else
			{
				sig_free(aSig);
			}
		}

		free(isFlag);

		if (rawSigGroup->noOfSig > 0)
		{
			free(rawSigGroup->sigList);
		}
		free(rawSigGroup);
	}

	




	return returnValue;
}


signatureGroupGroup *sig_createSignatureGroupFromLowerLevel(signatureGroupGroup *aSigGroupGroup, int threshold)
{
	signatureGroupGroup *returnValue;
	int i, j;

	signatureGroup *newSigGroup;
	signatureGroup *sigGroup1;
	signatureGroup *sigGroup2;
	int noOfSigType;

	signatureGroup **sigTypeList;


	returnValue = sigGroupGroup_new();

	noOfSigType = aSigGroupGroup->noOfGroup;
	sigTypeList = aSigGroupGroup->groupList;

	for (i = 0; i < noOfSigType; i++)
	{
		for (j = i+1; j < noOfSigType; j++)
		{
			sigGroup1 = sigTypeList[i];
			sigGroup2 = sigTypeList[j];
			
			newSigGroup = signatureRealSignatureForCombineAttrFromTwoSigGroup(sigGroup1, sigGroup2, threshold);
			if (newSigGroup != NULL)
			{
				sigGroupGroup_insert(returnValue, newSigGroup);
			}
		}
	}

	return returnValue;
}

void sigGroup_free(signatureGroup *aSigGroup)
{
	int noOfSig;

	int i;

	signature **sigList;

	sigList = aSigGroup->sigList;
	noOfSig = aSigGroup->noOfSig;

	if (noOfSig > 0)
	{
		for (i = 0; i < noOfSig; i++)
		{
			sig_free(sigList[i]);
		}
		free(sigList);
	}

	free(aSigGroup);
}



void sigGroupGroup_free(signatureGroupGroup *aSigGroupGroup)
{
	int noOfGroup;
	signatureGroup **sigGroupList;

	int i;

	noOfGroup = aSigGroupGroup->noOfGroup;
	sigGroupList = aSigGroupGroup->groupList;

	if (noOfGroup > 0)
	{
		for (i = 0; i < noOfGroup; i++)
		{
			sigGroup_free(sigGroupList[i]);
		}

		free(sigGroupList);
	}

	free(aSigGroupGroup);
}


signatureGroupGroupGroup *sigGroupGroupGroup_new()
{
	signatureGroupGroupGroup *returnValue;

	returnValue = (signatureGroupGroupGroup *) malloc(sizeof(signatureGroupGroupGroup));
	memset(returnValue, 0, sizeof(signatureGroupGroupGroup));

	return returnValue;
}

void sigGroupGroupGroup_insert(signatureGroupGroupGroup *aSigGroupGroupGroup, signatureGroupGroup *aSigGroupGroup)
{
	int noOfSize;
	signatureGroupGroup **sizeList;

	noOfSize = aSigGroupGroupGroup->noOfSize;
	sizeList = aSigGroupGroupGroup->sizeList;

	if (noOfSize == 0)
	{
		sizeList = (signatureGroupGroup **) malloc(sizeof(signatureGroupGroup *));
	}
	else
	{
		sizeList = (signatureGroupGroup **) realloc(sizeList, sizeof(signatureGroupGroup *)*(noOfSize+1));
	}
	sizeList[noOfSize] = aSigGroupGroup;

	aSigGroupGroupGroup->sizeList = sizeList;
	aSigGroupGroupGroup->noOfSize++;
}

signatureGroupGroupGroup *sig_findAllSignature(data *aData, int threshold)
{
	signatureGroupGroupGroup *returnValue;

	signatureGroupGroup *aSigGroupGroup;
	signatureGroupGroup *newSigGroupGroup;

	int size;

	returnValue = sigGroupGroupGroup_new();

	size = 1;
	printf("  Generating signature of size %d...\n", size);
	aSigGroupGroup = sig_createRealSignatureForMultipleAttr(aData, threshold);
	while (aSigGroupGroup->noOfGroup > 0)
	{
		sigGroupGroupGroup_insert(returnValue, aSigGroupGroup);

		size++;
		printf("  Generating signature of size %d...\n", size);
		newSigGroupGroup = sig_createSignatureGroupFromLowerLevel(aSigGroupGroup, threshold);

		aSigGroupGroup = newSigGroupGroup;
	}

	if (aSigGroupGroup->noOfGroup == 0)
	{
		sigGroupGroup_free(aSigGroupGroup);
	}

	return returnValue;
}

// -1: means no signature
int ***sig_findGlobalDistributionFromOrgTable(configure *config, signatureGroupGroupGroup *allSig, data *aData, int noOfSenIndex)
{
	int ***allSigMapping;

	int noOfSize;
	signatureGroupGroup **sizeList;

	signatureGroupGroup *aSigGroupGroup;
	signatureGroup **groupList;
	int noOfGroup;

	signatureGroup *aSigGroup;

	int i, j, k, l;

	int noOfSig;
	signature **sigList;

	int noOfMember;
	int *member;
	int memberID;

	signature *aSig;

	tuple *tupleList;
	tuple *aTuple;

	int noOfDim;
	int aSenValue;

	//int noOfSenValue;

	float *prob;

	int noOfTuple;

	int aSenIndex;

	noOfSize = allSig->noOfSize;
	sizeList = allSig->sizeList;

	tupleList = aData->tupleList;
	noOfDim = aData->noOfDim;

	//noOfSenValue = aData->dimSize[noOfDim-1];
	noOfTuple = aData->noOfTuple;


	allSigMapping = (int ***) malloc(sizeof(int **)*noOfTuple);
	for (l = 0; l < noOfTuple; l++)
	{
		allSigMapping[l] = (int **) malloc(sizeof(int *)*noOfSize);
		for (i = 0; i < noOfSize; i++)
		{
			aSigGroupGroup = sizeList[i];

			noOfGroup = aSigGroupGroup->noOfGroup;

			allSigMapping[l][i] = (int *) malloc(sizeof(int)*noOfGroup);

			for (k = 0; k < noOfGroup; k++)
			{
				allSigMapping[l][i][k] = -1;
			}
		}
	}

	for (i = 0; i < noOfSize; i++)
	{
		aSigGroupGroup = sizeList[i];

		noOfGroup = aSigGroupGroup->noOfGroup;
		groupList = aSigGroupGroup->groupList;
		
		for (j = 0; j < noOfGroup; j++)
		{
			aSigGroup = groupList[j];
			
			noOfSig = aSigGroup->noOfSig;
			sigList = aSigGroup->sigList;

			for (k = 0; k < noOfSig; k++)
			{
				aSig = sigList[k];

				

				if (aSig->prob == NULL)
				{
					//aSig->prob = (float *) malloc(sizeof(float)*noOfSenValue);
					aSig->prob = (float *) malloc(sizeof(float)*noOfSenIndex);
				}

				//memset(aSig->prob, 0, sizeof(float)*noOfSenValue);
				memset(aSig->prob, 0, sizeof(float)*noOfSenIndex);

				prob = aSig->prob;

				noOfMember = aSig->noOfMember;
				member = aSig->member;

				for (l = 0; l < noOfMember; l++)
				{
					memberID = member[l];

					aTuple = &(tupleList[memberID]);

					// value
					aSenValue = aTuple->content[noOfDim-1].value;
					// assume only two index
					if ((config->senStartNo <= aSenValue) && (aSenValue <= config->senEndNo))
					{
						aSenIndex = 1;
					}
					else
					{
						aSenIndex = 0;
					}
					
					// increment prob.
					//prob[aSenValue]++;
					prob[aSenIndex]++;

					// make index for tuple
					allSigMapping[memberID][i][j] = k;
				}

				// normalize
				//for (l = 0; l < noOfSenValue; l++)
				for (l = 0; l < noOfSenIndex; l++)
				{
					prob[l] /= noOfMember;
				}
			}
		}
	}

	return allSigMapping;
}

void sig_printStatistics(FILE *filePtr, signatureGroupGroupGroup *allSig)
{
	int noOfSize;
	signatureGroupGroup **sizeList;
	signatureGroupGroup *aSigGroupGroup;

	signatureGroup **groupList;
	signatureGroup *aSigGroup;

	int i, j;
	int noOfType;
	int noOfSig;

	noOfSize = allSig->noOfSize;
	sizeList = allSig->sizeList;

	fprintf(filePtr, "  no. of size = %d\n", noOfSize);
	for (i = 0; i < noOfSize; i++)
	{
		aSigGroupGroup = sizeList[i];

		noOfType = aSigGroupGroup->noOfGroup;
		groupList = aSigGroupGroup->groupList;
		fprintf(filePtr, "  Size = %d, no. of types = %d\n", i, noOfType);

		for (j = 0; j < noOfType; j++)
		{
			aSigGroup = groupList[j]; 
			noOfSig = aSigGroup->noOfSig;
			fprintf(filePtr, "     Type %d. no. of signatures = %d\n", j, noOfSig);
		}
	}

}

void sig_print(FILE *filePtr, signature *aSig, int noOfSenIndex)
{
	int noOfAttr;
	dimension *content;
	int *dimNo;
	float *prob;

	int i;

	noOfAttr = aSig->noOfAttr;
	content = aSig->content;
	
	dimNo = aSig->dimNo;
	prob = aSig->prob;

	fprintf(filePtr, "           no. of attr = %d\n", noOfAttr);
	fprintf(filePtr, "           content = (");
	for (i = 0; i < noOfAttr; i++)
	{
		fprintf(filePtr, "%d", content[i].value);
		if (i != noOfAttr-1)
		{
			fprintf(filePtr, ", ");
		}
	}
	fprintf(filePtr, ")\n");
	fprintf(filePtr, "           dimNo = (");
	for (i = 0; i < noOfAttr; i++)
	{
		fprintf(filePtr, "%d", dimNo[i]);
		if (i != noOfAttr-1)
		{
			fprintf(filePtr, ", ");
		}
	}
	fprintf(filePtr, ")\n");
	fprintf(filePtr, "           prob. = (");
	for (i = 0; i < noOfSenIndex; i++)
	{
		fprintf(filePtr, "%.10f", prob[i]);
		if (i != noOfSenIndex-1)
		{
			fprintf(filePtr, ", ");
		}
	}
	fprintf(filePtr, ")\n");
	 
}

void sig_printSignatureGroupGroupGroup(FILE *filePtr, signatureGroupGroupGroup *allSig, int noOfSenIndex)
{
	int noOfSize;
	signatureGroupGroup **sizeList;
	signatureGroupGroup *aSigGroupGroup;

	signatureGroup **groupList;
	signatureGroup *aSigGroup;
	signature **sigList;
	signature *aSig;

	int i, j, k;
	int noOfType;
	int noOfSig;


	noOfSize = allSig->noOfSize;
	sizeList = allSig->sizeList;

	fprintf(filePtr, "  no. of size = %d\n", noOfSize);
	for (i = 0; i < noOfSize; i++)
	{
		aSigGroupGroup = sizeList[i];

		noOfType = aSigGroupGroup->noOfGroup;
		groupList = aSigGroupGroup->groupList;
		fprintf(filePtr, "  Size = %d, no. of types = %d\n", i, noOfType);

		for (j = 0; j < noOfType; j++)
		{
			aSigGroup = groupList[j]; 
			noOfSig = aSigGroup->noOfSig;
			sigList = aSigGroup->sigList;
			fprintf(filePtr, "     Type %d. no. of signatures = %d\n", j, noOfSig);

			for (k = 0; k < noOfSig; k++)
			{
				fprintf(filePtr, "        Sig. %d. ", k);
				aSig = sigList[k];
				sig_print(filePtr, aSig, noOfSenIndex);
			}
		}
	}
}
